from __future__ import annotations

from functools import reduce
from operator import add

import numpy as np
import pytest
import torch
import torch.autograd.gradcheck
from torch.nn import Module

import pyqtorch as pyq
from pyqtorch.embed import ConcretizedCallable, Embedding, cos, log, sin, sqrt
from pyqtorch.primitives import Primitive
from pyqtorch.utils import ATOL_embedding


@pytest.mark.parametrize(
    "fn", ["sin", "cos", "log", "tanh", "tan", "sin", "sqrt", "square"]
)
def test_univariate(fn: str) -> None:
    results = []
    x = np.random.uniform(0, 1)
    for engine_name in ["jax", "torch", "numpy"]:
        native_call = ConcretizedCallable(fn, ["x"], {}, engine_name)
        native_result = native_call(
            {"x": (torch.tensor(x) if engine_name == "torch" else x)}
        )
        results.append(native_result.item())
    assert np.allclose(results[0], results[1], atol=ATOL_embedding) and np.allclose(
        results[0], results[2], ATOL_embedding
    )


@pytest.mark.parametrize("fn", ["mul", "add", "div", "sub"])
def test_multivariate(fn: str) -> None:
    results = []
    x = np.random.randn(1)
    y = np.random.randn(1)
    for engine_name in ["jax", "torch", "numpy"]:
        native_call = ConcretizedCallable(fn, ["x", "y"], {}, engine_name)
        native_result = native_call(
            {
                "x": torch.tensor(x) if engine_name == "torch" else x,
                "y": torch.tensor(y) if engine_name == "torch" else y,
            }
        )
        results.append(native_result.item())
    assert np.allclose(results[0], results[1], atol=ATOL_embedding) and np.allclose(
        results[0], results[2], atol=ATOL_embedding
    )


def test_embedding() -> None:
    x = np.random.uniform(0, 1)
    theta = np.random.uniform(0, 1)
    results = []
    for engine_name in ["jax", "torch", "numpy"]:
        v_params = ["theta"]
        f_params = ["x"]
        leaf0, native_call0 = "%0", ConcretizedCallable(
            "mul", ["x", "theta"], {}, engine_name
        )
        embedding = Embedding(
            v_params,
            f_params,
            var_to_call={leaf0: native_call0},
            engine_name=engine_name,
        )
        inputs = {
            "x": (torch.tensor(x) if engine_name == "torch" else x),
            "theta": (torch.tensor(theta) if engine_name == "torch" else theta),
        }
        eval_0 = embedding.var_to_call["%0"](inputs)
        results.append(eval_0.item())
    assert np.allclose(results[0], results[1], atol=ATOL_embedding) and np.allclose(
        results[0], results[2], atol=ATOL_embedding
    )


def test_reembedding() -> None:
    x = np.random.uniform(0, 1)
    theta = np.random.uniform(0, 1)
    t = np.random.uniform(0, 1)
    t_reembed = np.random.uniform(0, 1)
    results = []
    reembedded_results = []
    for engine_name in ["jax", "torch", "numpy"]:
        v_params = ["theta"]
        f_params = ["x"]
        tparam = "t"
        leaf0, native_call0 = "%0", ConcretizedCallable(
            "mul", ["x", "theta"], {}, engine_name
        )
        leaf1, native_call1 = "%1", ConcretizedCallable(
            "mul", ["t", "%0"], {}, engine_name
        )

        leaf2, native_call2 = "%2", ConcretizedCallable("sin", ["%1"], {}, engine_name)
        embedding = Embedding(
            v_params,
            f_params,
            var_to_call={leaf0: native_call0, leaf1: native_call1, leaf2: native_call2},
            tparam_name=tparam,
            engine_name=engine_name,
        )
        inputs = {
            "x": (torch.tensor(x) if engine_name == "torch" else x),
            "theta": (torch.tensor(theta) if engine_name == "torch" else theta),
            "t": (torch.tensor(t) if engine_name == "torch" else t),
        }
        all_params = embedding.embed_all(inputs)
        new_tparam_val = (
            torch.tensor(t_reembed) if engine_name == "torch" else t_reembed
        )
        reembedded_params = embedding.reembed_tparam(all_params, new_tparam_val)
        results.append(all_params["%2"].item())
        reembedded_results.append(reembedded_params["%2"].item())
    assert all([p in ["%1", "%2"] for p in embedding.tracked_vars])
    assert "%0" not in embedding.tracked_vars
    assert np.allclose(results[0], results[1], atol=ATOL_embedding) and np.allclose(
        results[0], results[2], atol=ATOL_embedding
    )
    assert np.allclose(
        reembedded_results[0], reembedded_results[1], atol=ATOL_embedding
    ) and np.allclose(reembedded_results[0], reembedded_results[2], atol=ATOL_embedding)


def test_sample_run_expectation_grads_with_embedding() -> None:
    name0, fn0 = "fn0", ConcretizedCallable("sin", ["x"])
    name1, fn1 = "fn1", ConcretizedCallable("mul", ["fn0", "y"])
    name2, fn2 = "fn2", ConcretizedCallable("mul", ["fn1", 2.0])
    name3, fn3 = "fn3", ConcretizedCallable("log", ["fn2"])
    embedding = pyq.Embedding(
        vparam_names=["x"],
        fparam_names=["y"],
        var_to_call={name0: fn0, name1: fn1, name2: fn2, name3: fn3},
    )
    rx = pyq.RX(0, param_name=name0)
    cry = pyq.CRY(0, 1, param_name=name1)
    phase = pyq.PHASE(1, param_name=name2)
    ry = pyq.RY(1, param_name=name3)
    cnot = pyq.CNOT(1, 2)
    ops = [rx, cry, phase, ry, cnot]
    n_qubits = 3
    circ = pyq.QuantumCircuit(n_qubits, ops)
    obs = pyq.Observable(pyq.Z(0))

    state = pyq.zero_state(n_qubits)

    x = torch.rand(1, requires_grad=True)
    y = torch.rand(1, requires_grad=True)

    values_ad = {"x": x, "y": y}

    wf = pyq.run(circ, state, values_ad, embedding)
    samples = pyq.sample(circ, state, values_ad, 100, embedding)
    exp_ad = pyq.expectation(
        circ, state, values_ad, obs, pyq.DiffMode.AD, embedding=embedding
    )
    assert torch.autograd.gradcheck(
        lambda x, y: pyq.expectation(
            circ, state, {"x": x, "y": y}, obs, pyq.DiffMode.AD, embedding=embedding
        ),
        (x, y),
        atol=1e-1,  # torch.autograd.gradcheck is very susceptible to small numerical errors
    )


@pytest.mark.parametrize("engine", ["torch", "jax", "numpy"])
def test_move_embedding(engine: str) -> None:
    name0, fn0 = "fn0", ConcretizedCallable("sin", ["x"], engine_name=engine)
    name1, fn1 = "fn1", ConcretizedCallable("mul", ["fn0", "y"], engine_name=engine)
    name2, fn2 = "fn2", ConcretizedCallable("mul", ["fn1", 2.0], engine_name=engine)
    name3, fn3 = "fn3", ConcretizedCallable("log", ["fn2"], engine_name=engine)
    embedding = pyq.Embedding(
        vparam_names=["x"],
        fparam_names=["y"],
        var_to_call={name0: fn0, name1: fn1, name2: fn2, name3: fn3},
        engine_name=engine,
    )
    embedding.to(device="cpu", dtype=torch.float32)
    embedding.to(device="cpu")
    embedding.to(dtype=torch.float32)


def test_reembedding_forward() -> None:

    class CustomReembed(pyq.Add):
        def __init__(self, operations: list[Module] | Primitive | pyq.Sequence):
            super().__init__(operations=operations)

        def forward(
            self,
            state: torch.Tensor,
            values: dict[str, torch.Tensor] | None = None,
            embedding: Embedding | None = None,
        ) -> torch.Tensor:
            values = values or dict()
            return reduce(
                add,
                [
                    self.run(state, values, embedding.reembed_tparam(values, val))  # type: ignore[union-attr, arg-type]
                    for val in torch.linspace(0.0, 10, 10)
                ],
            )

        def run(
            self,
            state: torch.Tensor,
            values: dict[str, torch.Tensor],
            embedding: Embedding | None = None,
        ) -> torch.Tensor:
            for op in self.operations:
                state = op(state, values)
            return state

    leaf0, native_call0 = "%0", ConcretizedCallable("sin", ["t"], {}, "torch")
    gen = pyq.Scale(pyq.Z(0), leaf0)
    custom = CustomReembed(gen)
    embed = pyq.Embedding(
        vparam_names=[],
        fparam_names=[],
        var_to_call={leaf0: native_call0},
        tparam_name="t",
    )
    wf = custom(state=pyq.zero_state(2), values={"t": torch.rand(1)}, embedding=embed)
    assert not torch.any(torch.isnan(wf))


def test_get_independent_args() -> None:
    expr: ConcretizedCallable = sqrt(sin("x")) + cos("r") * (1.0 / log("z") * "y")
    assert set(expr.independent_args) == {"x", "y", "z", "r"}


def test_sample_run_expectation_grads_with_embedding_observable(embedding_fixture):
    """
    Extended test for embedding with observables that have embedded parameters.

    This extends the existing test_sample_run_expectation_grads_with_embedding
    by testing against an observable with embedded parameters (Scale(Z(0), embedding)).

    Mathematical verification:
    ⟨O_embedded⟩ = ⟨ψ|f(x,y)·Z|ψ⟩ = f(x,y)·⟨ψ|Z|ψ⟩
    where f(x,y) = y * sin(x)
    """
    import torch
    import torch.autograd.gradcheck

    import pyqtorch as pyq
    from pyqtorch.composite import Scale

    embedding = embedding_fixture

    # Create circuit with embedded parameters
    rx = pyq.RX(0, param_name="sin_x")  # θ = sin(x)
    cry = pyq.CRY(0, 1, param_name="mul_sinx_y")  # θ = y*sin(x)
    phase = pyq.PHASE(1, param_name="mul_sinx_y")  # θ = y*sin(x)
    ry = pyq.RY(1, param_name="mul_sinx_y")  # θ = y*sin(x)
    cnot = pyq.CNOT(1, 2)

    ops = [rx, cry, phase, ry, cnot]
    n_qubits = 3
    circ = pyq.QuantumCircuit(n_qubits, ops)

    # Test observables
    obs_simple = pyq.Observable([pyq.Z(0)])  # Simple observable
    obs_embedded = pyq.Observable(
        [Scale(pyq.Z(0), param_name="mul_sinx_y")]
    )  # Embedded observable

    state = pyq.zero_state(n_qubits)

    # Parameter values
    x = torch.rand(1, requires_grad=True)
    y = torch.rand(1, requires_grad=True)
    values_ad = {"x": x, "y": y}

    # Test 1: Simple observable expectation
    exp_simple = pyq.expectation(
        circ, state, values_ad, obs_simple, pyq.DiffMode.AD, embedding=embedding
    )

    # Test 2: Embedded observable expectation
    exp_embedded = pyq.expectation(
        circ, state, values_ad, obs_embedded, pyq.DiffMode.AD, embedding=embedding
    )
    # y * sin(x)
    # Test 3: Manual verification
    embedded_values = embedding(values_ad)
    scaling_factor = embedded_values["mul_sinx_y"]

    # Manual calculation: f(x,y) * ⟨ψ|Z|ψ⟩
    exp_manual = scaling_factor * pyq.expectation(
        circ, state, embedded_values, obs_simple, pyq.DiffMode.AD
    )

    assert torch.allclose(
        exp_embedded, exp_manual, atol=1e-6
    ), "Embedded observable expectation should equal scaling_factor * simple_expectation"

    # Test 4: Gradient verification for embedded observable
    def loss_embedded(x_val, y_val):
        vals = {"x": x_val, "y": y_val}
        return pyq.expectation(
            circ, state, vals, obs_embedded, pyq.DiffMode.AD, embedding=embedding
        )

    # Verify gradients are computed correctly
    assert torch.autograd.gradcheck(
        loss_embedded, (x, y), atol=1e-1
    ), "Gradient check failed for embedded observable"

    # Test 5: Consistency check - observable evaluation equivalence
    # Method A: Observable with embedding
    exp_obs_embed = obs_embedded.expectation(state, values_ad, embedding)

    # Method B: Observable with pre-evaluated parameters
    exp_obs_eval = obs_embedded.expectation(state, embedded_values)

    assert torch.allclose(
        exp_obs_embed, exp_obs_eval, atol=1e-6
    ), "Observable expectation with embedding should match pre-evaluated expectation"

    # Test 6: Verify the mathematical relationship: ⟨f(x,y)·Z⟩ = f(x,y)·⟨Z⟩
    # Get the final state after circuit execution
    final_state = circ(state, values_ad, embedding)

    # Simple expectation ⟨Z⟩
    simple_exp_value = obs_simple.expectation(final_state)

    # Embedded expectation should equal scaling_factor * simple_expectation
    expected_embedded_value = scaling_factor * simple_exp_value
    actual_embedded_value = obs_embedded.expectation(final_state, embedded_values)

    assert torch.allclose(
        expected_embedded_value, actual_embedded_value, atol=1e-6
    ), "Mathematical relationship ⟨f(x,y)·Z⟩ = f(x,y)·⟨Z⟩ violated"

    print("✓ All embedding observable tests passed")


def test_embedding_full_support_consistency(embedding_fixture):
    """
    Test embedding consistency with full_support parameter in tensor methods.

    SIMPLIFIED VERSION: Uses only pyqtorch's built-in methods to avoid
    manual tensor manipulation issues.

    Mathematical verification:
    U_expanded(f(x,y)) = expand_operator(U_local(f(x,y)), local_support, full_support)
    """
    import torch

    import pyqtorch as pyq
    from pyqtorch.apply import apply_operator
    from pyqtorch.utils import expand_operator, random_state

    embedding = embedding_fixture

    # Test parameters
    batch_size = 2
    x = torch.rand(batch_size)
    y = torch.rand(batch_size)
    values = {"x": x, "y": y}

    # Create a gate on subset of qubits
    gate = pyq.CRX(0, 1, param_name="mul_sinx_y")  # Acts on qubits 0,1
    local_support = gate.qubit_support
    full_support = (0, 1, 2, 3)  # Expand to 4 qubits

    # Test 1: tensor() method with full_support vs manual expansion
    # Method A: Direct embedding with full_support
    tensor_full_embed = gate.tensor(values, embedding, full_support=full_support)

    # Method B: Local tensor + manual expansion
    tensor_local_embed = gate.tensor(values, embedding)
    tensor_full_manual = expand_operator(
        tensor_local_embed, local_support, full_support
    )

    assert torch.allclose(
        tensor_full_embed, tensor_full_manual, atol=1e-6
    ), "Full support tensor with embedding should match manual expansion"

    # Test 2: Consistency with pre-evaluation
    embedded_values = embedding(values)

    # Method C: Pre-evaluation with full_support
    tensor_full_eval = gate.tensor(embedded_values, full_support=full_support)

    assert torch.allclose(
        tensor_full_embed, tensor_full_eval, atol=1e-6
    ), "Full support tensor embedding should match pre-evaluated full support tensor"

    # Test 3: SIMPLIFIED state application using pyqtorch's apply_operator
    # Create state for full support
    state = random_state(len(full_support), batch_size)

    # Method A: Apply tensor with embedding using apply_operator
    result_embed = apply_operator(state.clone(), tensor_full_embed, full_support)

    # Method B: Apply manually expanded tensor using apply_operator
    result_manual = apply_operator(state.clone(), tensor_full_manual, full_support)

    assert torch.allclose(
        result_embed, result_manual, atol=1e-6
    ), "apply_operator results should be identical for embedding and manual expansion"

    # Test 4: End-to-end verification with circuit application
    # Create a circuit that uses the gate
    circuit = pyq.QuantumCircuit(len(full_support), [gate])

    # Method A: Circuit with embedding
    result_circuit_embed = circuit(state.clone(), values, embedding)

    # Method B: Circuit with pre-evaluated values
    result_circuit_eval = circuit(state.clone(), embedded_values)

    assert torch.allclose(
        result_circuit_embed, result_circuit_eval, atol=1e-6
    ), "Circuit execution with embedding should match pre-evaluated circuit"

    # Test 5: Verify expand_operator consistency with different supports
    # Test with minimal support vs full support
    minimal_gate = pyq.CRX(0, 1, param_name="mul_sinx_y")

    # Tensor on minimal support
    tensor_minimal = minimal_gate.tensor(values, embedding)

    # Expand to full support manually
    tensor_expanded = expand_operator(tensor_minimal, (0, 1), full_support)

    # Direct tensor on full support
    tensor_direct_full = minimal_gate.tensor(
        values, embedding, full_support=full_support
    )

    assert torch.allclose(
        tensor_expanded, tensor_direct_full, atol=1e-6
    ), "Manual expansion should match direct full_support tensor generation"

    print("✓ Full support embedding consistency tests passed")


def test_embedding_diagonal_operations(embedding_fixture):
    """
    Test embedding functionality with diagonal operations.

    Verifies that embeddings work correctly with diagonal gates and that
    the diagonal=True parameter produces consistent results.

    Mathematical verification for diagonal gates:
    diag(U(f(x,y))) should be consistent with dense(U(f(x,y)))
    """
    import torch

    import pyqtorch as pyq
    from pyqtorch.composite import Scale
    from pyqtorch.utils import random_state, todense_tensor

    embedding = embedding_fixture

    # Test parameters
    n_qubits = 3
    batch_size = 2
    x = torch.rand(batch_size)
    y = torch.rand(batch_size)
    values = {"x": x, "y": y}

    # Test diagonal gates (Z, PHASE, S, T, etc.)
    diagonal_gates = [
        pyq.RZ(0, param_name="mul_sinx_y"),
        pyq.PHASE(1, param_name="mul_sinx_y"),
        Scale(pyq.Z(2), "mul_sinx_y"),
    ]

    for gate in diagonal_gates:
        # Test 1: Verify gate is recognized as diagonal
        assert gate.is_diagonal, f"Gate {gate} should be diagonal"

        # Test 2: Dense tensor method
        tensor_dense_embed = gate.tensor(values, embedding, diagonal=False)
        tensor_dense_eval = gate.tensor(embedding(values), diagonal=False)

        assert torch.allclose(
            tensor_dense_embed, tensor_dense_eval, atol=1e-6
        ), f"Dense tensor embedding consistency failed for {gate}"

        # Test 3: Diagonal tensor method
        tensor_diag_embed = gate.tensor(values, embedding, diagonal=True)
        tensor_diag_eval = gate.tensor(embedding(values), diagonal=True)

        assert torch.allclose(
            tensor_diag_embed, tensor_diag_eval, atol=1e-6
        ), f"Diagonal tensor embedding consistency failed for {gate}"

        # Test 4: Diagonal vs Dense consistency
        dense_from_diag = todense_tensor(tensor_diag_embed)

        assert torch.allclose(
            tensor_dense_embed, dense_from_diag, atol=1e-6
        ), f"Diagonal and dense tensors should be equivalent for {gate}"

        # Test 5: Application to states
        state = random_state(n_qubits, batch_size)

        # Using dense tensor
        result_dense = gate(state.clone(), values, embedding)

        # Manual application with diagonal tensor (for verification)
        if hasattr(gate, "qubit_support"):
            from pyqtorch.apply import apply_operator

            result_manual = apply_operator(
                state.clone(), tensor_dense_embed, gate.qubit_support
            )

            assert torch.allclose(
                result_dense, result_manual, atol=1e-6
            ), f"Dense application should match manual tensor application for {gate}"

    print("✓ Diagonal operations embedding tests passed")



from __future__ import annotations

import random

import pytest
import torch
from helpers import calc_mat_vec_wavefunction, get_op_support, random_pauli_hamiltonian

from pyqtorch.composite import Add, Scale, Sequence
from pyqtorch.hamiltonians import GeneratorType, HamiltonianEvolution
from pyqtorch.primitives import (
    CNOT,
    CPHASE,
    CRX,
    CRY,
    CRZ,
    CSWAP,
    CY,
    CZ,
    OPS_2Q,
    OPS_3Q,
    OPS_DIAGONAL,
    OPS_DIAGONAL_PARAM,
    OPS_DIGITAL,
    OPS_PARAM,
    OPS_PARAM_2Q,
    SWAP,
    N,
    Parametric,
    Primitive,
    Projector,
    Toffoli,
)
from pyqtorch.utils import (
    ATOL,
    RTOL,
    density_mat,
    permute_basis,
    random_state,
    todense_tensor,
)

pi = torch.tensor(torch.pi)


@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("make_param", [True, False])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_primitive_tensor(
    n_qubits: int,
    make_param: bool,
    use_full_support: bool,
    batch_size: int,
    diagonal: bool,
) -> None:
    k_1q = 2 * n_qubits  # Number of 1-qubit terms
    k_2q = n_qubits**2  # Number of 2-qubit terms
    generator, param_list = random_pauli_hamiltonian(
        n_qubits, k_1q, k_2q, make_param, diagonal=diagonal
    )
    values = {param: torch.rand(1) for param in param_list}
    full_support = tuple(range(n_qubits)) if use_full_support else None
    generator_matrix = generator.tensor(
        values=values, full_support=full_support, diagonal=diagonal
    )
    generator_matrix = (
        generator_matrix.repeat((1, 1, batch_size))
        if len(generator_matrix.size()) == 3
        else generator_matrix.repeat((1, batch_size))
    )
    diag_op = diagonal and generator.is_diagonal
    primitive_op = Primitive(
        generator_matrix,
        qubit_support=full_support if use_full_support else generator.qubit_support,  # type: ignore[arg-type]
        diagonal=diagonal,
    )
    if diag_op:
        assert len(generator_matrix.size()) == 2
        assert primitive_op.is_diagonal
    else:
        assert len(generator_matrix.size()) == 3
        assert not primitive_op.is_diagonal
    assert torch.allclose(primitive_op.tensor(), generator_matrix, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_dm", [True, False])
@pytest.mark.parametrize("use_permute", [True, False])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
def test_digital_tensor(
    n_qubits: int,
    batch_size: int,
    use_full_support: bool,
    use_permute: bool,
    use_dm: bool,
) -> None:
    """
    Goes through all non-parametric gates and tests their application to a random state
    in comparison with the `tensor` method, either using just the qubit support of the gate
    or expanding its matrix to the maximum qubit support of the full circuit.
    """
    op: type[Primitive]
    for op in OPS_DIGITAL:
        supp = get_op_support(op, n_qubits)
        op_concrete = op(*supp)
        if op in OPS_DIAGONAL:
            assert op_concrete.is_diagonal
        psi_init = random_state(n_qubits, batch_size)
        if use_dm:
            psi_star = op_concrete(density_mat(psi_init))
        else:
            psi_star = op_concrete(psi_init)
        full_support = tuple(range(n_qubits)) if use_full_support else None
        psi_expected = calc_mat_vec_wavefunction(
            op_concrete, psi_init, full_support=full_support, use_permute=use_permute
        )
        if use_dm:
            psi_expected = density_mat(psi_expected)
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_dm", [True, False])
@pytest.mark.parametrize("use_permute", [True, False])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("dtype", [torch.complex64, torch.complex128])
def test_param_tensor(
    n_qubits: int,
    batch_size: int,
    dtype: torch.dtype,
    use_full_support: bool,
    use_permute: bool,
    use_dm: bool,
) -> None:
    """
    Goes through all parametric gates and tests their application to a random state
    in comparison with the `tensor` method, either using just the qubit support of the gate
    or expanding its matrix to the maximum qubit support of the full circuit.
    """
    op: type[Parametric]
    for op in OPS_PARAM:
        supp = get_op_support(op, n_qubits)
        params = [f"th{i}" for i in range(op.n_params)]
        op_concrete = op(*supp, *params).to(dtype=dtype)
        if op in OPS_DIAGONAL_PARAM:
            assert op_concrete.is_diagonal
        psi_init = random_state(n_qubits, dtype=dtype)
        values = {param: torch.rand(batch_size) for param in params}
        if use_dm:
            psi_star = op_concrete(density_mat(psi_init), values)
        else:
            psi_star = op_concrete(psi_init, values)
        assert psi_star.dtype == dtype
        full_support = tuple(range(n_qubits)) if use_full_support else None
        psi_expected = calc_mat_vec_wavefunction(
            op_concrete,
            psi_init,
            values=values,
            full_support=full_support,
            use_permute=use_permute,
        )
        if use_dm:
            psi_expected = density_mat(psi_expected)
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("compose", [Sequence, Add])
def test_sequence_tensor(
    n_qubits: int,
    batch_size: int,
    use_full_support: bool,
    compose: type[Sequence] | type[Add],
) -> None:
    op_list = []
    values = {}
    op: type[Primitive] | type[Parametric]
    """
    Builds a Sequence or Add composition of all possible gates on random qubit
    supports. Also assigns a Scale of a random value to the non-parametric gates.
    Tests the forward method (which goes through each gate individually) to the
    `tensor` method, which builds the full operator matrix and applies it.
    """
    for op in OPS_DIGITAL:
        supp = get_op_support(op, n_qubits)
        op_concrete = Scale(op(*supp), torch.rand(1))
        if op in OPS_DIAGONAL:
            assert op_concrete.is_diagonal
        op_list.append(op_concrete)
    for op in OPS_PARAM:
        supp = get_op_support(op, n_qubits)
        params = [f"{op.__name__}_th{i}" for i in range(op.n_params)]
        values.update({param: torch.rand(batch_size) for param in params})
        op_concrete = op(*supp, *params)
        if op in OPS_DIAGONAL_PARAM:
            assert op_concrete.is_diagonal
        op_list.append(op_concrete)
    random.shuffle(op_list)
    op_composite = compose(op_list)
    psi_init = random_state(n_qubits, batch_size)
    psi_star = op_composite(psi_init, values)
    full_support = tuple(range(n_qubits)) if use_full_support else None
    psi_expected = calc_mat_vec_wavefunction(
        op_composite, psi_init, values=values, full_support=full_support
    )
    assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("compose", [Sequence, Add])
def test_diagonal_sequence_tensor(
    n_qubits: int,
    batch_size: int,
    use_full_support: bool,
    compose: type[Sequence] | type[Add],
) -> None:
    op_list = []
    values = {}
    op: type[Primitive] | type[Parametric]
    for op in OPS_DIAGONAL.intersection(OPS_DIGITAL):
        supp = get_op_support(op, n_qubits)
        op_concrete = Scale(op(*supp), torch.rand(1))
        op_list.append(op_concrete)
    for op in OPS_DIAGONAL_PARAM.intersection(OPS_PARAM):
        supp = get_op_support(op, n_qubits)
        params = [f"{op.__name__}_th{i}" for i in range(op.n_params)]
        values.update({param: torch.rand(batch_size) for param in params})
        op_concrete = op(*supp, *params)
        op_list.append(op_concrete)
    random.shuffle(op_list)
    op_composite = compose(op_list)
    full_support = tuple(range(n_qubits)) if use_full_support else None
    assert op_composite.is_diagonal
    tensor_composite = op_composite.tensor(values, full_support=full_support)
    tensor_composite_diagonal = op_composite.tensor(
        values, full_support=full_support, diagonal=True
    )
    dense_diagonal = todense_tensor(tensor_composite_diagonal)
    assert torch.allclose(tensor_composite, dense_diagonal, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("n_proj", [1, 3])
@pytest.mark.parametrize("batch_size", [1, 5])
def test_projector_tensor(
    n_qubits: int, n_proj: int, batch_size: int, use_full_support: bool
) -> None:
    """
    Instantiates various random projectors on arbitrary qubit support
    and compares the forward method with directly applying the tensor.
    """
    iterations = 5
    for _ in range(iterations):
        rand_int_1 = random.randint(0, 2**n_proj - 1)
        rand_int_2 = random.randint(0, 2**n_proj - 1)
        bitstring1 = "{0:b}".format(rand_int_1).zfill(n_proj)
        bitstring2 = "{0:b}".format(rand_int_2).zfill(n_proj)
        supp = tuple(random.sample(range(n_qubits), n_proj))
        op = Projector(supp, bitstring1, bitstring2)
        psi_init = random_state(n_qubits, batch_size)
        psi_star = op(psi_init)
        full_support = tuple(range(n_qubits)) if use_full_support else None
        psi_expected = calc_mat_vec_wavefunction(
            op, psi_init, full_support=full_support
        )
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("operator", [N, CNOT, SWAP])
@pytest.mark.parametrize("use_full_support", [True, False])
def test_projector_vs_operator(
    n_qubits: int,
    operator: type[Primitive],
    use_full_support: bool,
) -> None:
    if operator == N:
        supp: tuple = (random.randint(0, n_qubits - 1),)
        op_concrete = N(*supp)
        projector = Projector(supp, "1", "1")
    if operator == CNOT:
        supp = tuple(random.sample(range(n_qubits), 2))
        op_concrete = CNOT(*supp)
        projector = Add(
            [
                Projector(supp, "00", "00"),
                Projector(supp, "01", "01"),
                Projector(supp, "10", "11"),
                Projector(supp, "11", "10"),
            ]
        )
    if operator == SWAP:
        supp = tuple(random.sample(range(n_qubits), 2))
        op_concrete = SWAP(*supp)
        projector = Add(
            [
                Projector(supp, "00", "00"),
                Projector(supp, "01", "10"),
                Projector(supp, "10", "01"),
                Projector(supp, "11", "11"),
            ]
        )
    full_support = tuple(range(n_qubits)) if use_full_support else None
    projector_mat = projector.tensor(full_support=full_support)
    operator_mat = op_concrete.tensor(full_support=full_support)
    assert torch.allclose(projector_mat, operator_mat, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("make_param", [True, False])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_hevo_pauli_tensor(
    n_qubits: int,
    make_param: bool,
    use_full_support: bool,
    batch_size: int,
    diagonal: bool,
) -> None:
    k_1q = 2 * n_qubits  # Number of 1-qubit terms
    k_2q = n_qubits**2  # Number of 2-qubit terms
    generator, param_list = random_pauli_hamiltonian(
        n_qubits, k_1q, k_2q, make_param, diagonal=diagonal
    )

    values = {param: torch.rand(batch_size) for param in param_list}
    psi_init = random_state(n_qubits, batch_size)
    full_support = tuple(range(n_qubits)) if use_full_support else None
    # Test the generator itself
    psi_star = generator(psi_init, values)
    psi_expected = calc_mat_vec_wavefunction(generator, psi_init, values, full_support)
    assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)

    # Test the hamiltonian evolution
    tparam = "t"
    operator = HamiltonianEvolution(generator, tparam)
    if diagonal:
        assert generator.is_diagonal
        if not make_param:
            assert operator.is_diagonal

    if make_param:
        assert operator.generator_type in (
            GeneratorType.PARAMETRIC_OPERATION,
            GeneratorType.PARAMETRIC_COMMUTING_SEQUENCE,
        )
    else:
        assert operator.generator_type in (
            GeneratorType.OPERATION,
            GeneratorType.COMMUTING_SEQUENCE,
        )
    values[tparam] = torch.rand(batch_size)
    psi_star = operator(psi_init, values)
    psi_expected = calc_mat_vec_wavefunction(operator, psi_init, values, full_support)
    assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("gen_qubits", [3, 4])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_hevo_tensor_tensor(
    gen_qubits: int,
    n_qubits: int,
    use_full_support: bool,
    batch_size: int,
    diagonal: bool,
) -> None:
    k_1q = 2 * gen_qubits  # Number of 1-qubit terms
    k_2q = gen_qubits**2  # Number of 2-qubit terms
    generator, _ = random_pauli_hamiltonian(gen_qubits, k_1q, k_2q, diagonal=diagonal)
    psi_init = random_state(n_qubits, batch_size)
    full_support = tuple(range(n_qubits)) if use_full_support else None
    # Test the hamiltonian evolution
    generator_matrix = generator.tensor()
    supp = tuple(random.sample(range(n_qubits), gen_qubits))
    tparam = "t"
    operator = HamiltonianEvolution(generator_matrix, tparam, supp)
    assert operator.generator_type == GeneratorType.TENSOR
    if diagonal:
        assert generator.is_diagonal
        assert operator.is_diagonal
    values = {tparam: torch.rand(batch_size)}
    psi_star = operator(psi_init, values)
    psi_expected = calc_mat_vec_wavefunction(operator, psi_init, values, full_support)
    assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("n_qubits", [3, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_permute_tensor(n_qubits: int, diagonal: bool) -> None:
    ops = OPS_2Q.union(OPS_3Q) if not diagonal else {CZ}
    for op in ops:
        supp, ordered_supp = get_op_support(op, n_qubits, get_ordered=True)

        op_concrete1 = op(*supp)
        op_concrete2 = op(*ordered_supp)

        mat1 = op_concrete1.tensor(diagonal=diagonal)
        mat2 = op_concrete2.tensor(diagonal=diagonal)

        perm = op_concrete1._qubit_support.qubits

        assert torch.allclose(
            mat1, permute_basis(mat2, perm, inv=True, diagonal=diagonal)
        )
        assert torch.allclose(mat2, permute_basis(mat1, perm, diagonal=diagonal))


@pytest.mark.parametrize("n_qubits", [3, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_permute_tensor_parametric(
    n_qubits: int, batch_size: int, diagonal: bool
) -> None:
    ops = (
        OPS_PARAM_2Q if not diagonal else OPS_PARAM_2Q.intersection(OPS_DIAGONAL_PARAM)
    )
    for op in ops:
        supp, ordered_supp = get_op_support(op, n_qubits, get_ordered=True)
        params = [f"{op.__name__}_th{i}" for i in range(op.n_params)]
        values = {param: torch.rand(batch_size) for param in params}

        op_concrete1 = op(*supp, *params)
        op_concrete2 = op(*ordered_supp, *params)

        mat1 = op_concrete1.tensor(values=values, diagonal=diagonal)
        mat2 = op_concrete2.tensor(values=values, diagonal=diagonal)

        perm = op_concrete1._qubit_support.qubits

        assert torch.allclose(
            mat1, permute_basis(mat2, perm, inv=True, diagonal=diagonal)
        )
        assert torch.allclose(mat2, permute_basis(mat1, perm, diagonal=diagonal))


def test_tensor_symmetries() -> None:
    assert not torch.allclose(CNOT(0, 1).tensor(), CNOT(1, 0).tensor())
    assert not torch.allclose(CY(0, 1).tensor(), CY(1, 0).tensor())
    assert not torch.allclose(CZ(0, 1).tensor(), CY(1, 0).tensor())
    assert not torch.allclose(CRX(0, 1, 1.0).tensor(), CRX(1, 0, 1.0).tensor())
    assert not torch.allclose(CRY(0, 1, 1.0).tensor(), CRY(1, 0, 1.0).tensor())
    assert not torch.allclose(CRZ(0, 1, 1.0).tensor(), CRZ(1, 0, 1.0).tensor())
    assert torch.allclose(CPHASE(0, 1, 1.0).tensor(), CPHASE(1, 0, 1.0).tensor())
    assert torch.allclose(SWAP(0, 1).tensor(), SWAP(1, 0).tensor())
    assert torch.allclose(CSWAP(0, (1, 2)).tensor(), CSWAP(0, (2, 1)).tensor())
    assert torch.allclose(CSWAP(1, (0, 2)).tensor(), CSWAP(1, (2, 0)).tensor())
    assert torch.allclose(CSWAP(2, (0, 1)).tensor(), CSWAP(2, (1, 0)).tensor())
    assert torch.allclose(Toffoli((0, 1), 2).tensor(), Toffoli((1, 0), 2).tensor())
    assert torch.allclose(Toffoli((0, 2), 1).tensor(), Toffoli((2, 0), 1).tensor())
    assert torch.allclose(Toffoli((1, 2), 0).tensor(), Toffoli((2, 1), 0).tensor())


def test_param_gate_embed(embedding_fixture):
    """
    Test embedding functionality with all parametric gates in OPS_PARAM.

    Tests both tensor and run methods over random states for all gates in OPS_PARAM.
    Verifies correctness by checking that methods return the same results when:
    1. Providing embedding in the methods vs
    2. Evaluating beforehand and applying gate with evaluated embedding

    Mathematical verification for each gate G:
    G(f(x,y)) |ψ⟩ ≡ G(θ_evaluated) |ψ⟩ where θ_evaluated = y * sin(x)

    COVERAGE: Tests all parametric gates except U gate.
    - ✅ Single-qubit gates: RX, RY, RZ, PHASE
    - ✅ Controlled gates: CPHASE, CRX, CRY, CRZ
    - ⚠️ U gate: Skipped due to embedding implementation limitations

    FIXED: Resolved numerical inconsistencies in controlled rotation gates (CRX, CRY, CRZ).
    The issue was caused by inconsistent qubit ordering in test's apply_operator call.
    Fix: Use gate.qubit_support instead of manual (control, target) tuple.
    """
    import random

    import torch

    from pyqtorch.primitives import (
        CPHASE,
        CRX,
        CRY,
        CRZ,
        OPS_PARAM,
        PHASE,
        RX,
        RY,
        RZ,
        U,
    )
    from pyqtorch.utils import ATOL, RTOL, random_state

    # Test parameters
    n_qubits = 4
    batch_size = 3
    embedding = embedding_fixture

    # Random parameter values
    x = torch.rand(batch_size, requires_grad=True)
    y = torch.rand(batch_size, requires_grad=True)
    values = {"x": x, "y": y}

    # Track which gates are tested vs skipped
    tested_gates = []
    skipped_gates = []

    print("🧪 Testing ALL parametric gates with embeddings...")
    print(f"📊 Testing with n_qubits={n_qubits}, batch_size={batch_size}")
    print("🔧 Embedding function: θ = y * sin(x)")

    # Test each parametric gate
    for gate_class in OPS_PARAM:

        # SKIP ONLY U GATE (embeddings not implemented)
        if gate_class == U:
            print(f"⚠️ Skipping {gate_class.__name__} - embeddings not implemented")
            skipped_gates.append(gate_class.__name__)
            continue

        print(f"\n🔍 Testing {gate_class.__name__}...")

        # Generate appropriate gate instances for ALL supported gates
        if gate_class in [RX, RY, RZ, PHASE]:
            # Single qubit gates
            target = random.randint(0, n_qubits - 1)
            gate = gate_class(target, param_name="mul_sinx_y")
            print(f"   📍 Single-qubit gate on target={target}")

        elif gate_class in [CPHASE, CRX, CRY, CRZ]:
            # Controlled gates (including rotation gates)
            control = random.randint(0, n_qubits - 1)
            target = random.choice([i for i in range(n_qubits) if i != control])
            gate = gate_class(control, target, param_name="mul_sinx_y")
            print(f"   🎯 Controlled gate: control={control}, target={target}")
            print(f"   📋 Gate qubit_support: {gate.qubit_support}")

        # Test all supported gates
        if gate_class in [RX, RY, RZ, PHASE, CPHASE, CRX, CRY, CRZ]:
            try:
                # Random initial state
                psi_init = random_state(n_qubits, batch_size)

                # Test 1: tensor() method consistency
                print("   🧮 Testing tensor() method...")
                # Method A: Direct embedding
                tensor_embed = gate.tensor(values, embedding)

                # Method B: Pre-evaluation
                embedded_values = embedding(values)
                tensor_eval = gate.tensor(embedded_values)

                assert torch.allclose(
                    tensor_embed, tensor_eval, rtol=RTOL, atol=ATOL
                ), f"Gate {gate_class.__name__} tensor method failed embedding consistency"
                print("      ✅ tensor() embedding consistency verified")

                # Test 2: run() method (forward pass) consistency
                print("   🏃 Testing run() method...")
                # Method A: Direct embedding
                result_embed = gate(psi_init.clone(), values, embedding)

                # Method B: Pre-evaluation
                result_eval = gate(psi_init.clone(), embedded_values)

                assert torch.allclose(
                    result_embed, result_eval, rtol=RTOL, atol=ATOL
                ), f"Gate {gate_class.__name__} run method failed embedding consistency"
                print("      ✅ run() embedding consistency verified")

                # Test 3: Verify tensor and run give same results
                print("   🔗 Testing tensor() vs run() consistency...")
                from pyqtorch.apply import apply_operator

                # Apply tensor method result manually using consistent qubit ordering
                tensor_result = apply_operator(
                    psi_init.clone(),
                    tensor_embed,
                    gate.qubit_support,  # Use gate's consistent qubit support
                )

                assert torch.allclose(
                    result_embed, tensor_result, rtol=RTOL, atol=ATOL
                ), f"Gate {gate_class.__name__} tensor and run methods inconsistent with embedding"
                print("      ✅ tensor() vs run() consistency verified")

                # Additional verification: Check numerical differences
                max_diff = torch.max(torch.abs(result_embed - tensor_result)).item()
                print(f"      📏 Max numerical difference: {max_diff:.2e}")

                print(f"✅ Gate {gate_class.__name__} PASSED all embedding tests")
                tested_gates.append(gate_class.__name__)

            except Exception as e:
                print(f"❌ Gate {gate_class.__name__} FAILED: {str(e)}")
                skipped_gates.append(gate_class.__name__)
                # Re-raise to see the full error for debugging
                raise

    # Comprehensive summary report
    print("\n" + "=" * 70)
    print("🧪 COMPLETE EMBEDDING TEST SUMMARY")
    print("=" * 70)
    print(f"✅ Successfully tested gates: {tested_gates}")
    print(f"⚠️ Skipped gates: {skipped_gates}")
    coverage_tested = len(tested_gates)
    coverage_total = len(tested_gates) + len(skipped_gates)
    coverage_percentage = 100 * coverage_tested / coverage_total
    print(
        f"📊 Coverage: {coverage_tested}/{coverage_total} gates tested "
        f"({coverage_percentage:.1f}%)"
    )
    # Detailed analysis
    single_qubit_gates = [
        gate for gate in tested_gates if gate in ["RX", "RY", "RZ", "PHASE"]
    ]
    controlled_gates = [
        gate for gate in tested_gates if gate in ["CPHASE", "CRX", "CRY", "CRZ"]
    ]

    print("\n📋 Gate type breakdown:")
    print(f"   🎯 Single-qubit rotation gates: {single_qubit_gates}")
    print(f"   🔗 Controlled gates: {controlled_gates}")

    # Assertions for test validation
    assert (
        len(tested_gates) > 0
    ), "No gates were successfully tested - this indicates a broader issue"

    # Verify we tested the core single-qubit rotation gates
    core_gates = ["RX", "RY", "RZ", "PHASE"]
    tested_core = [gate for gate in core_gates if gate in tested_gates]
    assert (
        len(tested_core) >= 3
    ), f"Should test at least 3 core rotation gates, only tested: {tested_core}"

    # NEW: Verify controlled rotation gates work
    controlled_rotation_gates = ["CRX", "CRY", "CRZ"]
    tested_controlled = [
        gate for gate in controlled_rotation_gates if gate in tested_gates
    ]

    if len(tested_controlled) == 3:
        print("🎉 BREAKTHROUGH: All controlled rotation gates now working!")
        print(f"   🔧 Fixed controlled gates: {tested_controlled}")
    elif len(tested_controlled) > 0:
        print(
            f"⚡ Partial success: {len(tested_controlled)}/3 controlled rotation gates working"
        )
        print(f"   ✅ Working: {tested_controlled}")
        failed_controlled = [
            gate for gate in controlled_rotation_gates if gate in skipped_gates
        ]
        print(f"   ❌ Failed: {failed_controlled}")
    else:
        print("⚠️ Controlled rotation gates still having issues")

    # Verify CPHASE still works
    assert "CPHASE" in tested_gates, "CPHASE should continue to work"

    print(
        f"\n🎯 RESULT: Successfully verified embedding consistency for {len(tested_gates)} gates"
    )
    print("✅ All tested gates passed embedding consistency checks!")



from __future__ import annotations

from collections import OrderedDict
from typing import Callable

import pytest
import torch
from helpers import calc_mat_vec_wavefunction, random_pauli_hamiltonian

import pyqtorch as pyq
from pyqtorch import RX, Add, ConcretizedCallable, HamiltonianEvolution, Scale, X
from pyqtorch.composite import Sequence
from pyqtorch.hamiltonians import GeneratorType
from pyqtorch.matrices import (
    DEFAULT_MATRIX_DTYPE,
    DEFAULT_REAL_DTYPE,
    IMAT,
    XMAT,
    ZMAT,
)
from pyqtorch.noise import AnalogDepolarizing
from pyqtorch.utils import (
    ATOL,
    RTOL,
    DensityMatrix,
    SolverType,
    density_mat,
    is_normalized,
    operator_kron,
    overlap,
    product_state,
    random_state,
)

pi = torch.tensor(torch.pi)


def Hamiltonian(batch_size: int = 1) -> torch.Tensor:
    sigmaz = torch.diag(torch.tensor([1.0, -1.0], dtype=DEFAULT_MATRIX_DTYPE))
    Hbase = torch.kron(sigmaz, sigmaz)
    H = torch.kron(Hbase, Hbase)
    if batch_size == 1:
        return H
    elif batch_size == 2:
        return torch.stack((H, H.conj()), dim=2)
    else:
        raise NotImplementedError


def Hamiltonian_general(n_qubits: int = 2, batch_size: int = 1) -> torch.Tensor:
    H_batch = torch.zeros(
        (2**n_qubits, 2**n_qubits, batch_size), dtype=DEFAULT_MATRIX_DTYPE
    )
    for i in range(batch_size):
        H_0 = torch.randn((2**n_qubits, 2**n_qubits), dtype=DEFAULT_MATRIX_DTYPE)
        H = (H_0 + torch.conj(H_0.transpose(0, 1))).to(DEFAULT_MATRIX_DTYPE)
        H_batch[..., i] = H
    return H_batch


def Hamiltonian_diag(n_qubits: int = 2, batch_size: int = 1) -> torch.Tensor:
    H_batch = torch.zeros(
        (2**n_qubits, 2**n_qubits, batch_size), dtype=DEFAULT_MATRIX_DTYPE
    )
    for i in range(batch_size):
        H_0 = torch.randn((2**n_qubits, 2**n_qubits), dtype=DEFAULT_MATRIX_DTYPE)
        H = H_0 + torch.conj(H_0.transpose(0, 1))
        get_diag = torch.diag(H)
        H_diag = torch.diag(get_diag)
        H_batch[..., i] = H_diag
    return H_batch


@pytest.mark.flaky(max_runs=5)
@pytest.mark.parametrize("n_qubits, batch_size", [(2, 1), (4, 2)])
def test_hamevo_general(n_qubits: int, batch_size: int) -> None:
    H = Hamiltonian_general(n_qubits, batch_size)
    t_evo = torch.rand(1, dtype=DEFAULT_REAL_DTYPE)
    hamevo = pyq.HamiltonianEvolution(H, t_evo, tuple([i for i in range(n_qubits)]))
    psi = random_state(n_qubits, batch_size)
    psi_star = hamevo(psi)
    assert is_normalized(psi_star, atol=ATOL)


@pytest.mark.flaky(max_runs=5)
def test_hamevo_single() -> None:
    n_qubits = 4
    H = Hamiltonian(1)
    t_evo = torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE)
    hamevo = pyq.HamiltonianEvolution(H, t_evo, tuple([i for i in range(n_qubits)]))
    psi = pyq.uniform_state(n_qubits)
    psi_star = hamevo(psi)
    result = overlap(psi_star, psi)
    assert torch.isclose(
        result, torch.tensor([0.5], dtype=torch.float64), rtol=RTOL, atol=ATOL
    )


@pytest.mark.flaky(max_runs=5)
def test_hamevo_batch() -> None:
    n_qubits = 4
    batch_size = 2
    H = Hamiltonian(batch_size)
    t_evo = torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE)
    hamevo = pyq.HamiltonianEvolution(H, t_evo, tuple([i for i in range(n_qubits)]))
    psi = pyq.uniform_state(n_qubits, batch_size)
    psi_star = hamevo(psi)
    result = overlap(psi_star, psi)
    assert torch.allclose(
        result, torch.tensor([0.5, 0.5], dtype=torch.float64), rtol=RTOL, atol=ATOL
    )


@pytest.mark.parametrize(
    "H, t_evo, target, batch_size",
    [
        (  # batchsize 1 | 1
            Hamiltonian(1),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5], dtype=torch.float64),
            1,
        ),
        (  # batchsize 1 | 4
            Hamiltonian(1),
            torch.tensor([torch.pi / 4, 0.0, torch.pi / 2, torch.pi]),
            torch.tensor([0.5, 1.0, 0.0, 1.0], dtype=torch.float64),
            4,
        ),
        (  # batchsize 2 | 1
            Hamiltonian(2),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5, 0.5], dtype=torch.float64),
            2,
        ),
        (  # batchsize 2 | 2
            Hamiltonian(2),
            torch.tensor([torch.pi / 4, torch.pi]),
            torch.tensor([0.5, 1.0], dtype=torch.float64),
            2,
        ),
    ],
)
def test_hamiltonianevolution_with_types(
    H: torch.Tensor,
    t_evo: torch.Tensor,
    target: torch.Tensor,
    batch_size: int,
) -> None:
    n_qubits = 4
    hamevo = pyq.HamiltonianEvolution(H, t_evo, tuple([i for i in range(n_qubits)]))
    assert len(hamevo._cache_hamiltonian_evo) == 0
    psi = pyq.uniform_state(n_qubits)
    psi_star = hamevo(psi)
    result = overlap(psi_star, psi)
    assert result.size() == (batch_size,)
    assert torch.allclose(result, target, rtol=RTOL, atol=ATOL)

    # test cached
    assert len(hamevo._cache_hamiltonian_evo) == 1

    psi_star = hamevo(psi)
    assert len(hamevo._cache_hamiltonian_evo) == 1
    result = overlap(psi_star, psi)
    # check value has not changed
    assert torch.allclose(result, target, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize(
    "H, t_evo, expected_state",
    [
        (
            Hamiltonian(1),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5], dtype=torch.float64),
        ),
        (
            Hamiltonian(1),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5, 0.5], dtype=torch.float64),
        ),
        (
            Hamiltonian(1),
            torch.tensor([torch.pi / 4, torch.pi]),
            torch.tensor([0.5, 1.0], dtype=torch.float64),
        ),
        # with batchdim
        (
            Hamiltonian(2),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5, 0.5], dtype=torch.float64),
        ),
        # with batchdim in position 0
        (
            Hamiltonian(2).transpose(0, 2),
            torch.tensor([torch.pi / 4], dtype=DEFAULT_MATRIX_DTYPE),
            torch.tensor([0.5, 0.5], dtype=torch.float64),
        ),
    ],
)
def test_symbol_hamevo(
    H: torch.Tensor,
    t_evo: torch.Tensor,
    expected_state: torch.Tensor,
) -> None:
    symbol = "h"
    n_qubits = 4
    hamevo = pyq.HamiltonianEvolution(
        symbol, t_evo, tuple([i for i in range(n_qubits)])
    )
    assert hamevo.generator_type == GeneratorType.SYMBOL
    psi = pyq.uniform_state(n_qubits)
    psi_star = hamevo(psi, {symbol: H})
    state = overlap(psi_star, psi)
    assert torch.allclose(state, expected_state, rtol=RTOL, atol=ATOL)


def test_hamevo_fixed_tensor_result() -> None:
    hermitian_matrix = torch.tensor([[2.0, 1.0], [1.0, 3.0]], dtype=torch.complex128)
    hamevo = pyq.HamiltonianEvolution(
        generator=hermitian_matrix, time=torch.tensor([1.0j]), qubit_support=(0,)
    )
    assert hamevo.generator_type == GeneratorType.TENSOR
    expected_evo_result = torch.tensor(
        [[[13.1815 + 0.0j], [14.8839 + 0.0j]], [[14.8839 + 0.0j], [28.0655 + 0.0j]]],
        dtype=torch.complex128,
    )
    assert torch.allclose(hamevo.tensor(), expected_evo_result, atol=1.0e-4)


@pytest.mark.parametrize(
    "state_fn", [pyq.random_state, pyq.zero_state, pyq.uniform_state]
)
def test_parametric_phase_hamevo(
    state_fn: Callable, batch_size: int = 1, n_qubits: int = 1
) -> None:
    target = 0
    state = state_fn(n_qubits, batch_size=batch_size)
    phi = torch.rand(1, dtype=DEFAULT_MATRIX_DTYPE)
    H = (ZMAT - IMAT) / 2
    hamevo = pyq.HamiltonianEvolution(H, phi, (target,))
    assert hamevo.generator_type == GeneratorType.TENSOR
    phase = pyq.PHASE(target, "phi")
    assert torch.allclose(phase(state, {"phi": phi}), hamevo(state))


def test_hamevo_endianness() -> None:
    t = torch.ones(1)
    h = torch.tensor(
        [
            [0.9701, 0.0000, 0.7078, 0.0000],
            [0.0000, 0.9701, 0.0000, 0.7078],
            [0.4594, 0.0000, 0.9207, 0.0000],
            [0.0000, 0.4594, 0.0000, 0.9207],
        ],
        dtype=torch.complex128,
    )
    iszero = torch.tensor([False, True, False, True])
    op = pyq.HamiltonianEvolution(qubit_support=(0, 1), generator=h, time=t)
    st = op(pyq.zero_state(2)).flatten()
    assert torch.allclose(
        st[iszero], torch.zeros(1, dtype=DEFAULT_MATRIX_DTYPE), rtol=RTOL, atol=ATOL
    )

    h = torch.tensor(
        [
            [0.9701, 0.7078, 0.0000, 0.0000],
            [0.4594, 0.9207, 0.0000, 0.0000],
            [0.0000, 0.0000, 0.9701, 0.7078],
            [0.0000, 0.0000, 0.4594, 0.9207],
        ],
        dtype=torch.complex128,
    )
    iszero = torch.tensor([False, False, True, True])
    op = pyq.HamiltonianEvolution(qubit_support=(0, 1), generator=h, time=t)
    st = op(pyq.zero_state(2)).flatten()
    assert torch.allclose(
        st[iszero], torch.zeros(1, dtype=DEFAULT_MATRIX_DTYPE), rtol=RTOL, atol=ATOL
    )


def test_hamevo_endianness_cnot() -> None:
    n_qubits = 2
    state_10 = product_state("10")

    gen = -0.5 * operator_kron((IMAT - ZMAT).unsqueeze(-1), (IMAT - XMAT).unsqueeze(-1))
    hamiltonian_evolution = pyq.HamiltonianEvolution(
        generator=gen,
        time=torch.tensor([torch.pi / 2.0]),
        qubit_support=tuple(range(n_qubits)),
    )
    wf_hamevo = hamiltonian_evolution(state_10)

    cnot_op = pyq.CNOT(0, 1)
    wf_cnot = cnot_op(state_10)
    assert torch.allclose(wf_cnot, wf_hamevo, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("duration", [torch.rand(1), "duration"])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("use_sparse", [True, False])
@pytest.mark.parametrize("ode_solver", [SolverType.DP5_SE, SolverType.KRYLOV_SE])
def test_timedependent(
    tparam: str,
    param_y: float,
    duration: float,
    batch_size: int,
    use_sparse: bool,
    n_steps: int,
    torch_hamiltonian: Callable,
    hamevo_generator: Sequence,
    sin: tuple,
    sq: tuple,
    ode_solver: SolverType,
) -> None:

    psi_start = random_state(2, batch_size)

    dur_val = duration if isinstance(duration, torch.Tensor) else torch.rand(1)

    # simulate with time-dependent solver
    t_points = torch.linspace(0, dur_val[0], n_steps)

    psi_solver = pyq.sesolve(
        torch_hamiltonian,
        psi_start.reshape(-1, batch_size),
        t_points,
        ode_solver,
        options={"use_sparse": use_sparse},
    ).states[-1]

    # simulate with HamiltonianEvolution
    embedding = pyq.Embedding(
        tparam_name=tparam,
        var_to_call={sin[0]: sin[1], sq[0]: sq[1]},
    )
    hamiltonian_evolution = pyq.HamiltonianEvolution(
        generator=hamevo_generator,
        time=tparam,
        duration=dur_val,
        steps=n_steps,
        solver=ode_solver,
        use_sparse=use_sparse,
    )
    values = {"y": param_y, "duration": dur_val}
    psi_hamevo = hamiltonian_evolution(
        state=psi_start, values=values, embedding=embedding
    ).reshape(-1, batch_size)

    assert torch.allclose(psi_solver, psi_hamevo, rtol=RTOL, atol=1.0e-3)


@pytest.mark.parametrize("duration", [torch.rand(1), "duration"])
@pytest.mark.parametrize(
    "batchsize",
    [
        1,
        3,
    ],
)
def test_timedependent_with_noise(
    tparam: str,
    param_y: float,
    duration: float,
    batchsize: int,
    n_steps: int,
    torch_hamiltonian: Callable,
    hamevo_generator: Sequence,
    sin: tuple,
    sq: tuple,
) -> None:

    psi_start = density_mat(random_state(2, batchsize))
    dur_val = duration if isinstance(duration, torch.Tensor) else torch.rand(1)

    # simulate with time-dependent solver
    t_points = torch.linspace(0, dur_val[0], n_steps)

    # Define jump operators
    # Note that we squeeze to remove the batch dimension
    noise = AnalogDepolarizing(error_param=0.1, qubit_support=0)
    list_ops = noise._noise_operators(full_support=(0, 1))
    solver = SolverType.DP5_ME
    psi_solver = pyq.mesolve(
        torch_hamiltonian, psi_start, list_ops, t_points, solver
    ).states[-1]

    # simulate with HamiltonianEvolution
    embedding = pyq.Embedding(
        tparam_name=tparam,
        var_to_call={sin[0]: sin[1], sq[0]: sq[1]},
    )
    hamiltonian_evolution = pyq.HamiltonianEvolution(
        generator=hamevo_generator,
        time=tparam,
        duration=duration,
        steps=n_steps,
        solver=solver,
        noise=list_ops,
    )
    values = {"y": param_y, "duration": dur_val}
    psi_hamevo = hamiltonian_evolution(
        state=psi_start, values=values, embedding=embedding
    )
    assert isinstance(psi_hamevo, DensityMatrix)
    assert torch.allclose(psi_solver, psi_hamevo, rtol=RTOL, atol=1.0e-3)


def test_error_noise_qubit_support(
    tparam: str,
    duration: float,
    n_steps: int,
    hamevo_generator: Sequence,
):
    solver = SolverType.DP5_ME
    with pytest.raises(ValueError):
        noise = AnalogDepolarizing(error_param=0.1, qubit_support=3)
        hamevo = pyq.HamiltonianEvolution(
            generator=hamevo_generator,
            time=tparam,
            duration=duration,
            steps=n_steps,
            solver=solver,
            noise=noise,
        )


@pytest.mark.parametrize("n_qubits", [2, 4, 6])
@pytest.mark.parametrize("batch_size", [1, 2])
@pytest.mark.parametrize("make_param", [True, False])
@pytest.mark.parametrize("commuting_terms", [False, True])
def test_hamevo_parametric_gen(
    n_qubits: int, batch_size: int, make_param: bool, commuting_terms: bool
) -> None:
    k_1q = 2 * n_qubits  # Number of 1-qubit terms
    k_2q = n_qubits**2  # Number of 2-qubit terms
    generator, param_list = random_pauli_hamiltonian(
        n_qubits,
        k_1q,
        k_2q,
        make_param=make_param,
        p_param=1.0,
        commuting_terms=commuting_terms,
    )
    tparam = "t"

    param_list.append("t")

    hamevo = pyq.HamiltonianEvolution(generator, tparam, cache_length=2)

    if make_param:
        assert hamevo.generator_type in (
            GeneratorType.PARAMETRIC_OPERATION,
            GeneratorType.PARAMETRIC_COMMUTING_SEQUENCE,
        )
    else:
        assert hamevo.generator_type in (
            GeneratorType.OPERATION,
            GeneratorType.COMMUTING_SEQUENCE,
        )
    assert len(hamevo._cache_hamiltonian_evo) == 0

    def apply_hamevo_and_compare_expected(psi, values):
        psi_star = hamevo(psi, values)
        psi_expected = calc_mat_vec_wavefunction(hamevo, psi, values)
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)

    values = {param: torch.rand(batch_size) for param in param_list}

    psi = random_state(n_qubits)
    apply_hamevo_and_compare_expected(psi, values)

    # test cached
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 1
    apply_hamevo_and_compare_expected(psi, values)
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 1

    # test caching new value
    for param in param_list:
        values[param] += 0.1

    apply_hamevo_and_compare_expected(psi, values)
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 2

    # changing input state should not change the cache
    psi = random_state(n_qubits)
    apply_hamevo_and_compare_expected(psi, values)
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 2

    # test limit cache
    previous_cache_keys = hamevo._cache_hamiltonian_evo.keys()

    for param in param_list:
        values[param] += 0.1

    values_cache_key = str(OrderedDict(values))
    if not commuting_terms:
        assert values_cache_key not in previous_cache_keys

    apply_hamevo_and_compare_expected(psi, values)
    if not commuting_terms:
        assert len(hamevo._cache_hamiltonian_evo) == 2
        assert values_cache_key in previous_cache_keys


@pytest.mark.parametrize(
    "generator, time_param, result",
    [
        (RX(0, "x"), "x", True),
        (RX(1, 0.5), "y", False),
        (RX(0, "x"), "y", False),
        (RX(0, "x"), torch.tensor(0.5), False),
        (RX(0, torch.tensor(0.5)), torch.tensor(0.5), False),
        (Scale(X(1), "y"), "y", True),
        (Scale(X(1), 0.2), "x", False),
        (
            Add(
                [Scale(X(1), ConcretizedCallable("mul", ["y", "x"])), Scale(X(1), "z")]
            ),
            "x",
            True,
        ),
        (
            Add(
                [Scale(X(1), ConcretizedCallable("add", ["y", "x"])), Scale(X(1), "z")]
            ),
            "t",
            False,
        ),
    ],
)
def test_hamevo_is_time_dependent_generator(generator, time_param, result) -> None:
    hamevo = HamiltonianEvolution(generator, time_param)
    assert hamevo.is_time_dependent == result


@pytest.mark.parametrize("batch_size", [1, 2])
def test_hamiltonian_evolution_parameter_reembedding_complex(batch_size: int) -> None:
    """
    Test parameter re-embedding with complex nested expressions.

    This test creates a more complex parameter dependency chain similar to
    the expressions mentioned in pasqal-io/qadence#492:

    Mathematical structure:
    base_expr = x * theta
    trig_expr = sin(base_expr * t)
    final_coeff = (omega * y * trig_expr) + (alpha * cos(t))
    H(t) = final_coeff * σₓ

    Verifies that re-embedding correctly handles:
    1. Nested ConcretizedCallable dependencies
    2. Multiple levels of parameter composition
    3. Mixed time-dependent and time-independent parameters
    """
    tparam = "t"

    # Create nested parameter expressions using BINARY operations only
    # Layer 1: base computation
    base_expr, base_fn = "base_expr", ConcretizedCallable("mul", ["x", "theta"])

    # Layer 2: time-dependent expressions
    base_times_t, base_times_t_fn = (
        "base_times_t",
        ConcretizedCallable("mul", ["base_expr", tparam]),
    )
    sin_base_t, sin_base_t_fn = "sin_base_t", ConcretizedCallable(
        "sin", ["base_times_t"]
    )
    cos_t, cos_t_fn = "cos_t", ConcretizedCallable("cos", [tparam])

    # Layer 3: binary multiplication chains
    omega_y, omega_y_fn = "omega_y", ConcretizedCallable("mul", ["omega", "y"])
    term1, term1_fn = "term1", ConcretizedCallable("mul", ["omega_y", "sin_base_t"])
    term2, term2_fn = "term2", ConcretizedCallable("mul", ["alpha", "cos_t"])
    final_coeff, final_coeff_fn = "final_coeff", ConcretizedCallable(
        "add", ["term1", "term2"]
    )

    # Create generator: H(t) = final_coeff * X(0)
    generator = Scale(pyq.X(0), final_coeff)

    # Create embedding with all parameter mappings
    embedding = pyq.Embedding(
        vparam_names=["x", "theta", "y"],  # Variational parameters
        fparam_names=["omega", "alpha"],  # Feature parameters
        var_to_call={
            base_expr: base_fn,
            base_times_t: base_times_t_fn,
            sin_base_t: sin_base_t_fn,
            cos_t: cos_t_fn,
            omega_y: omega_y_fn,
            term1: term1_fn,
            term2: term2_fn,
            final_coeff: final_coeff_fn,
        },
        tparam_name=tparam,
    )

    # Create HamiltonianEvolution
    hamevo = pyq.HamiltonianEvolution(
        generator=generator, time=tparam, qubit_support=(0,)
    )

    # Test parameters - INCLUDE time parameter from the start
    values = {
        "x": torch.tensor(1.5),
        "theta": torch.tensor(0.8),
        "y": torch.tensor(2.0),
        "omega": torch.tensor(1.2),
        "alpha": torch.tensor(0.7),
        tparam: torch.tensor(0.0),
    }

    # Test different time values
    test_times = [0.1, 0.5, 1.0, 2.0]

    for t_val in test_times:
        # Full embedding approach
        values_full = values.copy()
        values_full[tparam] = torch.tensor(t_val)
        embedded_full = embedding.embed_all(values_full.copy())

        # Re-embedding approach
        embedded_base = embedding.embed_all(values.copy())
        embedded_reembed = embedding.reembed_tparam(
            embedded_base.copy(), torch.tensor(t_val)
        )

        # Get evolution operators
        op_full = hamevo.tensor(values=embedded_full, embedding=None)
        op_reembed = hamevo.tensor(values=embedded_reembed, embedding=None)

        # Verify equivalence
        assert torch.allclose(
            op_full, op_reembed, rtol=1e-12, atol=1e-14
        ), f"Complex re-embedding failed for t={t_val}"

        # Verify that all time-dependent parameters are tracked
        expected_tracked = {base_times_t, sin_base_t, cos_t, term1, term2, final_coeff}
        actual_tracked = set(embedding.tracked_vars)
        assert expected_tracked <= actual_tracked, (
            f"Missing tracked vars: expected {expected_tracked}, "
            f"got {actual_tracked}"
        )


def test_hamiltonian_evolution_parameter_reembedding_state_evolution() -> None:
    """
    Test that parameter re-embedding produces identical quantum state evolution.

    This test verifies the complete quantum evolution process, not just the
    tensor construction, ensuring that re-embedding doesn't introduce any
    phase errors or unitary violations in the time evolution.
    """
    tparam = "t"
    n_qubits = 2

    # Create time-dependent generator: H(t) = (ω·t)·cos(ω·t)·(X⊗I + I⊗Z)
    # Using binary operations: (omega * t) * cos(omega * t)
    omega_t, omega_t_fn = "omega_t", ConcretizedCallable("mul", ["omega", tparam])
    cos_omega_t, cos_omega_t_fn = "cos_omega_t", ConcretizedCallable("cos", ["omega_t"])
    omega_times_t, omega_times_t_fn = "omega_times_t", ConcretizedCallable(
        "mul", ["omega", tparam]
    )
    coeff, coeff_fn = "coeff", ConcretizedCallable(
        "mul", ["omega_times_t", "cos_omega_t"]
    )

    generator = Scale(Add([pyq.X(0), pyq.Z(1)]), coeff)  # X⊗I + I⊗Z

    embedding = pyq.Embedding(
        fparam_names=["omega"],
        var_to_call={
            omega_t: omega_t_fn,
            cos_omega_t: cos_omega_t_fn,
            omega_times_t: omega_times_t_fn,
            coeff: coeff_fn,
        },
        tparam_name=tparam,
    )

    hamevo = pyq.HamiltonianEvolution(
        generator=generator, time=tparam, qubit_support=tuple(range(n_qubits))
    )

    # Initial state |1⟩⊗|0⟩
    initial_state = pyq.product_state("10")  # |1⟩⊗|0⟩
    omega_val = torch.tensor(1.5)

    # Test multiple evolution times
    evolution_times = [0.1, 0.5, 1.0]

    for t_val in evolution_times:
        # Method 1: Full embedding - INCLUDE time parameter from start
        values_full = {"omega": omega_val, tparam: torch.tensor(t_val)}
        embedded_full = embedding.embed_all(values_full.copy())
        final_state_full = hamevo(initial_state, values=embedded_full, embedding=None)

        # Method 2: Re-embedding - INCLUDE time parameter from start
        values_base = {"omega": omega_val, tparam: torch.tensor(0.0)}
        embedded_base = embedding.embed_all(values_base.copy())
        embedded_reembed = embedding.reembed_tparam(
            embedded_base.copy(), torch.tensor(t_val)
        )
        final_state_reembed = hamevo(
            initial_state, values=embedded_reembed, embedding=None
        )

        # Verify state equivalence
        assert torch.allclose(
            final_state_full, final_state_reembed, rtol=1e-12, atol=1e-14
        ), f"State evolution differs for t={t_val}"

        # Verify normalization is preserved
        norm_full = torch.sum(torch.abs(final_state_full.flatten()) ** 2)
        norm_reembed = torch.sum(torch.abs(final_state_reembed.flatten()) ** 2)
        # Fix dtype mismatch: use same dtype as the computed norm
        expected_norm = torch.tensor(1.0, dtype=norm_full.dtype)
        assert torch.allclose(norm_full, expected_norm, atol=1e-12)
        assert torch.allclose(norm_reembed, expected_norm, atol=1e-12)


def test_hamiltonian_evolution_parameter_reembedding_edge_cases() -> None:
    """
    Test edge cases for parameter re-embedding to ensure robustness.

    Edge cases tested:
    1. Parameters that don't depend on time (should not be in tracked_vars)
    2. Zero time evolution
    3. Parameters with constant values
    4. Empty parameter dependencies
    """
    tparam = "t"

    # Case 1: Mixed time-dependent and time-independent parameters
    sin_t, sin_t_fn = "sin_t", ConcretizedCallable("sin", [tparam])
    const_expr, const_expr_fn = "const_expr", ConcretizedCallable("mul", ["a", "b"])

    generator = Add(
        [
            Scale(pyq.X(0), sin_t),  # Time-dependent
            Scale(pyq.Z(0), const_expr),  # Time-independent
        ]
    )

    embedding = pyq.Embedding(
        fparam_names=["a", "b"],
        var_to_call={
            sin_t: sin_t_fn,
            const_expr: const_expr_fn,
        },
        tparam_name=tparam,
    )

    hamevo = pyq.HamiltonianEvolution(generator, tparam, qubit_support=(0,))

    values = {"a": torch.tensor(1.0), "b": torch.tensor(2.0)}

    # Initialize embedding to trigger tracking analysis
    values_init = values.copy()
    values_init[tparam] = torch.tensor(0.0)
    embedded_base = embedding.embed_all(values_init)

    # Verify tracking: only sin_t should be tracked
    assert sin_t in embedding.tracked_vars, "sin_t should be tracked (depends on time)"
    assert (
        const_expr not in embedding.tracked_vars
    ), "const_expr should not be tracked (time-independent)"
    # Case 2: Zero time evolution should work
    embedded_zero = embedding.reembed_tparam(embedded_base.copy(), torch.tensor(0.0))
    op_zero = hamevo.tensor(values=embedded_zero, embedding=None)

    # Should be close to identity for small times
    assert torch.allclose(
        torch.diagonal(op_zero, dim1=0, dim2=1).squeeze(),
        torch.tensor([1.0 + 0j, 1.0 + 0j], dtype=torch.complex128),
        rtol=1e-10,
        atol=1e-12,
    )

    # Case 3: Large time values should not cause numerical instability
    large_t = torch.tensor(100.0)
    embedded_large = embedding.reembed_tparam(embedded_base.copy(), large_t)
    op_large = hamevo.tensor(values=embedded_large, embedding=None)

    # Should still be unitary
    op_dagger = torch.conj(op_large.transpose(0, 1))
    identity_check = torch.matmul(op_large.squeeze(), op_dagger.squeeze())
    expected_identity = torch.eye(2, dtype=torch.complex128)
    assert torch.allclose(identity_check, expected_identity, rtol=1e-10, atol=1e-12)


@pytest.mark.parametrize("n_time_steps", [5, 10])
def test_hamiltonian_evolution_parameter_reembedding_performance_verification(
    n_time_steps: int,
) -> None:
    """
    Test that re-embedding maintains mathematical equivalence across many time steps.

    This test simulates a scenario where the same Hamiltonian is evaluated at many
    different time points, which is typical in time-dependent evolution. It verifies
    that re-embedding produces identical results to full embedding at every step.
    """
    tparam = "t"

    # Create a moderately complex time-dependent expression
    # H(t) = [ω₁·sin(ωt) + ω₂·cos(ωt)]·X + [β·t²]·Z
    omega_t, omega_t_fn = "omega_t", ConcretizedCallable("mul", ["omega", tparam])
    sin_omega_t, sin_omega_t_fn = "sin_omega_t", ConcretizedCallable("sin", ["omega_t"])
    cos_omega_t, cos_omega_t_fn = "cos_omega_t", ConcretizedCallable("cos", ["omega_t"])
    t_squared, t_squared_fn = "t_squared", ConcretizedCallable("mul", [tparam, tparam])

    term1, term1_fn = "term1", ConcretizedCallable("mul", ["omega1", "sin_omega_t"])
    term2, term2_fn = "term2", ConcretizedCallable("mul", ["omega2", "cos_omega_t"])
    x_coeff, x_coeff_fn = "x_coeff", ConcretizedCallable("add", ["term1", "term2"])
    z_coeff, z_coeff_fn = "z_coeff", ConcretizedCallable("mul", ["beta", "t_squared"])

    generator = Add(
        [
            Scale(pyq.X(0), x_coeff),
            Scale(pyq.Z(0), z_coeff),
        ]
    )

    embedding = pyq.Embedding(
        fparam_names=["omega", "omega1", "omega2", "beta"],
        var_to_call={
            omega_t: omega_t_fn,
            sin_omega_t: sin_omega_t_fn,
            cos_omega_t: cos_omega_t_fn,
            t_squared: t_squared_fn,
            term1: term1_fn,
            term2: term2_fn,
            x_coeff: x_coeff_fn,
            z_coeff: z_coeff_fn,
        },
        tparam_name=tparam,
    )

    hamevo = pyq.HamiltonianEvolution(generator, tparam, qubit_support=(0,))

    # Parameters
    params = {
        "omega": torch.tensor(2.0),
        "omega1": torch.tensor(1.5),
        "omega2": torch.tensor(0.8),
        "beta": torch.tensor(0.3),
    }

    # Generate time points
    time_points = torch.linspace(0.0, 2.0, n_time_steps)

    # Initialize base embedding
    params_base = params.copy()
    params_base[tparam] = torch.tensor(0.0)
    embedded_base = embedding.embed_all(params_base.copy())

    max_difference = 0.0

    for t_val in time_points:
        # Full embedding
        params_full = params.copy()
        params_full[tparam] = t_val
        embedded_full = embedding.embed_all(params_full.copy())
        op_full = hamevo.tensor(values=embedded_full, embedding=None)

        # Re-embedding
        embedded_reembed = embedding.reembed_tparam(embedded_base.copy(), t_val)
        op_reembed = hamevo.tensor(values=embedded_reembed, embedding=None)

        # Track maximum difference
        diff = torch.max(torch.abs(op_full - op_reembed))
        max_difference = max(max_difference, diff.item())

        # Verify equivalence at each step
        assert torch.allclose(
            op_full, op_reembed, rtol=1e-12, atol=1e-14
        ), f"Difference found at t={t_val}: {diff}"

    # Ensure overall precision is maintained
    assert (
        max_difference < 1e-13
    ), f"Maximum difference {max_difference} exceeds tolerance"

    # Verify that all expected parameters are tracked
    expected_tracked = {
        omega_t,
        sin_omega_t,
        cos_omega_t,
        t_squared,
        term1,
        term2,
        x_coeff,
        z_coeff,
    }
    actual_tracked = set(embedding.tracked_vars)
    assert (
        expected_tracked <= actual_tracked
    ), f"Missing tracked parameters: {expected_tracked - actual_tracked}"


@pytest.mark.parametrize("batch_size", [1, 3])
@pytest.mark.parametrize("n_qubits", [2, 3])
def test_hamiltonian_evolution_parameter_reembedding_basic(
    batch_size: int, n_qubits: int
) -> None:
    """
    Test basic parameter re-embedding functionality in time-dependent HamiltonianEvolution.

    This test verifies that reembed_tparam produces identical results to full embedding
    for a simple time-dependent generator using sin(t) and t^2 terms.

    Mathematical verification:
    H(t) = ω[y·sin(t)·σₓ⊗I + x·t²·I⊗σᵧ]

    The test ensures that:
    embedding.embed_all() ≡ embedding.reembed_tparam()
    for different time values, up to numerical precision.
    """
    # Create time-dependent parameter expressions using ConcretizedCallable
    tparam = "t"
    sin_t, sin_t_fn = "sin_t", ConcretizedCallable("sin", [tparam])
    t_sq, t_sq_fn = "t_sq", ConcretizedCallable("mul", [tparam, tparam])

    # Create a complex generator with nested dependencies
    # H(t) = ω[y·sin(t)·X(0) + x·t²·Y(1)]
    omega = 2.5
    generator = Scale(
        Add(
            [
                Scale(Scale(pyq.X(0), sin_t), "y"),  # y·sin(t)·X(0)
                Scale(Scale(pyq.Y(1), t_sq), "x"),  # x·t²·Y(1)
            ]
        ),
        omega,
    )

    # Create embedding with time parameter tracking
    embedding = pyq.Embedding(
        vparam_names=["x", "y"],  # Variational parameters
        fparam_names=[],  # No feature parameters
        var_to_call={sin_t: sin_t_fn, t_sq: t_sq_fn},  # Time-dependent mappings
        tparam_name=tparam,  # Track time parameter
    )

    # Create HamiltonianEvolution
    hamevo = pyq.HamiltonianEvolution(
        generator=generator, time=tparam, qubit_support=tuple(range(n_qubits))
    )

    # Test parameters - INCLUDE time parameter from the start
    x_val = torch.rand(batch_size, requires_grad=True)
    y_val = torch.rand(batch_size, requires_grad=True)
    base_values = {"x": x_val, "y": y_val, tparam: torch.tensor(0.0)}  # Include time

    # Test multiple time values to ensure robustness
    test_times = [0.0, 0.5, 1.0, 1.5708, 3.14159]  # Include 0, π/2, π

    for t_val in test_times:
        values_full = base_values.copy()
        values_full[tparam] = torch.tensor(t_val)

        # Method 1: Full embedding (recalculates everything)
        embedded_full = embedding.embed_all(values_full.copy())
        evolved_op_full = hamevo.tensor(values=embedded_full, embedding=None)

        # Method 2: Re-embedding (only recalculates time-dependent parameters)
        embedded_base = embedding.embed_all(base_values.copy())
        embedded_reembed = embedding.reembed_tparam(
            embedded_base.copy(), torch.tensor(t_val)
        )
        evolved_op_reembed = hamevo.tensor(values=embedded_reembed, embedding=None)

        # Verify mathematical equivalence
        assert torch.allclose(
            evolved_op_full, evolved_op_reembed, rtol=1e-12, atol=1e-14
        ), (
            f"Re-embedding failed for t={t_val}. "
            f"Max diff: {torch.max(torch.abs(evolved_op_full - evolved_op_reembed))}"
        )
        # Verify that tracked variables are correctly identified
        expected_tracked = {sin_t, t_sq}  # Both depend on time
        assert (
            set(embedding.tracked_vars) == expected_tracked
        ), f"Expected tracked vars {expected_tracked}, got {set(embedding.tracked_vars)}"



from __future__ import annotations

import torch
from torch import Tensor

from pyqtorch.time_dependent.integrators.krylov import KrylovIntegrator


class Krylov(KrylovIntegrator):
    """Krylov subspace method for solving Schrodinger equation."""

    def integrate(self, t0: float, t1: float, y: Tensor) -> Tensor:
        dt = t1 - t0
        y = self.step(dt, t1, y)
        return y

    def step(
        self,
        dt: float,
        t: float,
        state: Tensor,
    ) -> Tensor:

        # ensure Krylov solver is passed only 2D tensors
        if state.ndim != 2:
            raise ValueError(
                f"Krylov solver expects a 2D state tensor with a batch dimension"
                f" at index 1 (expected state.ndim == 2; got state.ndim == {state.ndim})."
            )

        # move trailing batch dimension to leading batch dimension
        state = state.T.unsqueeze(-1)

        def exponentiate() -> tuple[torch.Tensor, bool]:
            # approximate next iteration by modifying T, and unmodifying
            T[:, i - 1, i] = 0
            T[:, i + 1, i] = 1
            exp = torch.linalg.matrix_exp(-1j * dt * T[:, : i + 2, : i + 2].clone())
            T[:, i - 1, i] = T[:, i, i - 1]
            T[:, i + 1, i] = 0

            # compute errors
            e1 = abs(exp[:, i, 0])
            e2 = abs(exp[:, i + 1, 0]) * n

            # compute batch convergence error
            error = (e1 * e2) / (e1 - e2)
            use_e2_idx = e1 > 10 * e2
            use_e1_idx = e2 > e1
            error[use_e2_idx] = e2[use_e2_idx]
            error[use_e1_idx] = e1[use_e1_idx]

            # set convergence criteria based on maximum error in batch
            converged = error.max() < self.options.exp_tolerance
            return exp[:, :, 0], converged

        lanczos_vectors = [state]
        T = torch.zeros(
            (state.shape[0], self.options.max_krylov + 1, self.options.max_krylov + 1),
            dtype=state.dtype,
        )

        # step 0 of the loop
        v = torch.matmul(self.H(t), state)
        a = torch.matmul(v.conj().mT, state)
        a_scalar = a.squeeze((1, 2))

        n = torch.linalg.vector_norm(v.squeeze(2), dim=1)
        T[:, 0, 0] = a_scalar
        v = v - a * state

        for i in range(1, self.options.max_krylov):

            # this block should not be executed in step 0
            b = torch.linalg.vector_norm(v.squeeze(2), dim=1).view(-1, 1, 1)
            b_scalar = b.squeeze((1, 2))

            if b_scalar.max() < self.options.norm_tolerance:
                exp = torch.linalg.matrix_exp(-1j * dt * T[:, :i, :i])
                weights = exp[:, :, 0]
                converged = True
                break

            T[:, i, i - 1] = b_scalar
            T[:, i - 1, i] = b_scalar
            state = v / b

            lanczos_vectors.append(state)
            weights, converged = exponentiate()

            if converged:
                break

            v = torch.matmul(self.H(t), state)
            a = torch.matmul(v.conj().mT, state)

            n = torch.linalg.vector_norm(v.squeeze(2), dim=1)
            T[:, i, i] = a_scalar
            v = v - a * lanczos_vectors[i] - b * lanczos_vectors[i - 1]

        if not converged:
            raise RecursionError(
                "Exponentiation algorithm did not converge \
                to precision in allotted number of steps."
            )

        lanczos_vector_stack = torch.stack(lanczos_vectors, dim=1)
        n_lanczos = lanczos_vector_stack.shape[1]

        # take weighted combination of lanczos vectors and
        # move the batch dimension (b) back to the end
        result = torch.einsum(
            "bijk, bi -> jb", lanczos_vector_stack, weights[:, :n_lanczos]
        )

        return result
